function [Abar, Bbar, Cbar, Dbar, T] = transf2nf(A,B,C,D,mu)
%transf2nf Transforms the LTI system xdot = A*x + B*u + D*Delta, y = C*x
% into the proposed observability canonical form for strongly observable
% systems with unknown inputs using the state transformation z = T^(-1)*x.
%   A... dynamic matrix of the given system
%   B... control input matrix of the given system
%   C... output matrix of the given system
%   D... unknown input matrix of the given system
%   mu... vector mu = [mu_1 mu_2 ... mu_m] containing the orders of the
%         subsystems. The sum of all mu_j must equal the system order.
%
%   Abar... dynamic matrix of the system in observability canonical form
%   Bbar... control input matrix of the system in observability canonical form
%   Cbar... output matrix of the system in observability canonical form
%   Dbar... unknown input matrix of the system in observability canonical form
%   T... transformation matrix (z = T^(-1)*x)

n = length(A);

%% Reorder outputs (rows of C) such that mu_j is in descending order
[mu_sort,order] = sort(mu, 'descend');
C_sort = C(order, :);

% remove outputs (rows of C_sort) which are not used (mu_j = 0)
C_sort = C_sort(mu_sort ~= 0,:);
mu_sort = mu_sort(mu_sort ~= 0);
m_sort = length(mu_sort); % number of outputs which are used


%% Step 1: calculate O_R
O_R = zeros(n);

if (sum(mu_sort) ~= n)
    disp('Error: The sum of all mu_j has to equal n!')
    return;
end

k = 1;
for j = 1:m_sort
    for i = 1:mu_sort(j)
        O_R(k,:) = C_sort(j,:)*A^(i-1);
        k = k+1;
    end
end

if rank(O_R) ~= n
    disp('Error: The matrix O_R does not have full rank for the given choice of mu!')
end

%% Step 2: calculate the column vectors t_(mu_1), t_(mu_1 + mu_2),..., t_(n)  
t = zeros(n,m_sort);

for j = 1:m_sort
    t(:,j) = O_R\[zeros(sum(mu_sort(1:j))-1,1) ; 1; zeros(n-sum(mu_sort(1:j)),1)];
end

%% Step 3: calculate the coefficients beta_(j,k,l)
beta = {}; % store coefficients in a cell array

for j = 1:(m_sort-1)
    
    % Step 3.1: construct H^(j)
    H_j = {};
    
    for r = (j+1):m_sort
        for s = (j+1):m_sort
            
            last_col_of_H_j_r_s = zeros(mu_sort(j)-mu_sort(r),1);
            
            if ((mu_sort(j) == mu_sort(r)) || (mu_sort(j) == mu_sort(s)))
                H_j{r-j,s-j} = [];
            else
            
                for k = 1:length(last_col_of_H_j_r_s)
                    last_col_of_H_j_r_s(k) =  C_sort(r,:)*A^(mu_sort(j)-1-k)*t(:,s);
                end

                last_row_of_H_j_r_p = [zeros(1, mu_sort(j) - mu_sort(s)-1), last_col_of_H_j_r_s(end)];

                H_j{r-j,s-j} = rot90(toeplitz(fliplr(last_row_of_H_j_r_p), flipud(last_col_of_H_j_r_s))', 2);
            end
        end
        
    end
    
    % Step 3.2: construct w^(j)
    w_j = {};
    
    for r = (j+1):m_sort
        
        w_j_r = zeros(mu_sort(j)-mu_sort(r),1);
        
        for k = 1:length(w_j_r)
            w_j_r(k) =  C_sort(r,:)*A^(mu_sort(j)-k)*t(:,j);
        end
        
        w_j{r-j,1} = w_j_r;
    end
    
    % Step 3.3: calculate beta_(j,k,l)
    H_j = cell2mat(H_j);
    w_j = cell2mat(w_j);
    beta_j = H_j\w_j;
    
    for r = (j+1):m_sort
        beta{r-j,j} = [zeros(1,mu_sort(r)-1),beta_j(1:(mu_sort(j)-mu_sort(r)))'];
        beta_j = beta_j((mu_sort(j)-mu_sort(r)+1):end);
    end
end

% Step 4: construct transformation matrix T
T =  zeros(n);

for j = 1:m_sort
    
    for i = 0:(mu_sort(j)-1)
        T(:,sum(mu_sort(1:j))-i) = A^(i)*t(:,j);
        
        for r = (j+1):m_sort
            for q = 1:i
                beta_j = cell2mat(beta(1:(m_sort-j),j));
                
                T(:,sum(mu_sort(1:j))-i) = T(:,sum(mu_sort(1:j))-i) - beta_j(r-j,mu_sort(j)-q)*A^(i-q)*t(:,r);
            end
        end
    end
end

Abar = T\A*T;
Bbar = T\B;
Cbar = C*T;
Dbar = T\D;

end

